@@ -14,73 +14,107 @@ from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
 
 from .exceptions import check_error, exception
 
-def gendevice(devtype, host, mac, name=None, cloud=None):
-    devices = {
-        sp1: [0],
-        sp2: [0x2711,  # SP2
-              0x2719, 0x7919, 0x271a, 0x791a,  # Honeywell SP2
-              0x2720,  # SPMini
-              0x753e,  # SP3
-              0x7D00,  # OEM branded SP3
-              0x947a, 0x9479,  # SP3S
-              0x2728,  # SPMini2
-              0x2733, 0x273e,  # OEM branded SPMini
-              0x7530, 0x7546, 0x7918,  # OEM branded SPMini2
-              0x7D0D,  # TMall OEM SPMini3
-              0x2736  # SPMiniPlus
-              ],
-        rm: [0x2712,  # RM2
-             0x2737,  # RM Mini
-             0x273d,  # RM Pro Phicomm
-             0x2783,  # RM2 Home Plus
-             0x277c,  # RM2 Home Plus GDT
-             0x272a,  # RM2 Pro Plus
-             0x2787,  # RM2 Pro Plus2
-             0x279d,  # RM2 Pro Plus3
-             0x27a9,  # RM2 Pro Plus_300
-             0x278b,  # RM2 Pro Plus BL
-             0x2797,  # RM2 Pro Plus HYC
-             0x27a1,  # RM2 Pro Plus R1
-             0x27a6,  # RM2 Pro PP
-             0x278f,  # RM Mini Shate
-             0x27c2,  # RM Mini 3
-             0x27d1,  # new RM Mini3
-             0x27de  # RM Mini 3 (C)
-             ],
-        rm4: [0x51da,  # RM4 Mini
-              0x5f36,  # RM Mini 3
-              0x6026,  # RM4 Pro
-              0x6070,  # RM4c Mini
-              0x610e,  # RM4 Mini
-              0x610f,  # RM4c
-              0x62bc,  # RM4 Mini
-              0x62be  # RM4c Mini
-              ],
-        a1: [0x2714],  # A1
-        mp1: [0x4EB5,  # MP1
-              0x4EF7  # Honyar oem mp1
-              ],
-        hysen: [0x4EAD],  # Hysen controller
-        S1C: [0x2722],  # S1 (SmartOne Alarm Kit)
-        dooya: [0x4E4D],  # Dooya DT360E (DOOYA_CURTAIN_V2)
-        bg1: [0x51E3], # BG Electrical Smart Power Socket
-        lb1 : [0x60c8]   # RGB Smart Bulb
+
+def get_devices():
+    return {
+        0x0000: (sp1, "SP1", "Broadlink"),
+
+        0x2711: (sp2, "SP2", "Broadlink"),
+        0x2719: (sp2, "SP2-compatible", "Honeywell"),
+        0x271a: (sp2, "SP2-compatible", "Honeywell"),
+        0x2720: (sp2, "SP mini", "Broadlink"),
+        0x2728: (sp2, "SP2-compatible", "URANT"),
+        0x2733: (sp2, "SP3", "Broadlink"),
+        0x2736: (sp2, "SP mini+", "Broadlink"),
+        0x273e: (sp2, "SP mini", "Broadlink"),
+        0x7530: (sp2, "SP2", "Broadlink (OEM)"),
+        0x753e: (sp2, "SP mini 3", "Broadlink"),
+        0X7544: (sp2, "SP2-CL", "Broadlink"),
+        0x7546: (sp2, "SP2-UK/BR/IN", "Broadlink (OEM)"),
+        0x7547: (sp2, "SC1", "Broadlink"),
+        0x7918: (sp2, "SP2", "Broadlink (OEM)"),
+        0x7919: (sp2, "SP2-compatible", "Honeywell"),
+        0x791a: (sp2, "SP2-compatible", "Honeywell"),
+        0x7d00: (sp2, "SP3-EU", "Broadlink (OEM)"),
+        0x7d0d: (sp2, "SP mini 3", "Broadlink (OEM)"),
+        0x9479: (sp2, "SP3S-US", "Broadlink"),
+        0x947a: (sp2, "SP3S-EU", "Broadlink"),
+
+        0x2712: (rm, "RM pro/pro+", "Broadlink"),
+        0x272a: (rm, "RM pro", "Broadlink"),
+        0x2737: (rm, "RM mini 3", "Broadlink"),
+        0x273d: (rm, "RM pro", "Broadlink"),
+        0x277c: (rm, "RM home", "Broadlink"),
+        0x2783: (rm, "RM home", "Broadlink"),
+        0x2787: (rm, "RM pro", "Broadlink"),
+        0x278b: (rm, "RM plus", "Broadlink"),
+        0x278f: (rm, "RM mini", "Broadlink"),
+        0x2797: (rm, "RM pro+", "Broadlink"),
+        0x279d: (rm, "RM pro+", "Broadlink"),
+        0x27a1: (rm, "RM plus", "Broadlink"),
+        0x27a6: (rm, "RM plus", "Broadlink"),
+        0x27a9: (rm, "RM pro+", "Broadlink"),
+        0x27c2: (rm, "RM mini 3", "Broadlink"),
+        0x27d1: (rm, "RM mini 3", "Broadlink"),
+        0x27de: (rm, "RM mini 3", "Broadlink"),
+
+        0x51da: (rm4, "RM4 mini", "Broadlink"),
+        0x5f36: (rm4, "RM mini 3", "Broadlink"),
+        0x6026: (rm4, "RM4 pro", "Broadlink"),
+        0x6070: (rm4, "RM4C mini", "Broadlink"),
+        0x610e: (rm4, "RM4 mini", "Broadlink"),
+        0x610f: (rm4, "RM4C mini", "Broadlink"),
+        0x61a2: (rm4, "RM4 pro", "Broadlink"),
+        0x62bc: (rm4, "RM4 mini", "Broadlink"),
+        0x62be: (rm4, "RM4C mini", "Broadlink"),
+
+        0x2714: (a1, "e-Sensor", "Broadlink"),
+
+        0x4eb5: (mp1, "MP1-1K4S", "Broadlink"),
+        0x4ef7: (mp1, "MP1-1K4S", "Broadlink (OEM)"),
+        0x4f65: (mp1, "MP1-1K3S2U", "Broadlink"),
+
+        0x5043: (lb1, "SB800TD", "Broadlink (OEM)"),
+        0x504e: (lb1, "LB1", "Broadlink"),
+        0x60c7: (lb1, "LB1", "Broadlink"),
+        0x60c8: (lb1, "LB1", "Broadlink"),
+        0x6112: (lb1, "LB1", "Broadlink"),
+
+        0x2722: (S1C, "S2KIT", "Broadlink"),
+
+        0x4ead: (hysen, "HY02B05H", "Hysen"),
+
+        0x4e4d: (dooya, "DT360E-45/20", "Dooya"),
+
+        0x51e3: (bg1, "BG800/BG900", "BG Electrical"),
     }
 
-    # Look for the class associated to devtype in devices
-    [device_class] = [dev for dev in devices if devtype in devices[dev]] or [None]
-    if device_class is None:
-        return device(host, mac, devtype, name=name, cloud=cloud)
-    return device_class(host, mac, devtype, name=name, cloud=cloud)
+
+def gendevice(dev_type, host, mac, name=None, is_locked=None):
+    """Generate a device."""
+    try:
+        dev_class, model, manufacturer = get_devices()[dev_type]
+
+    except KeyError:
+        return device(host, mac, dev_type, name=name, is_locked=is_locked)
+
+    return dev_class(
+        host,
+        mac,
+        dev_type,
+        name=name,
+        model=model,
+        manufacturer=manufacturer,
+        is_locked=is_locked,
+    )
 
 
 def discover(timeout=None, local_ip_address=None, discover_ip_address='255.255.255.255'):
     if local_ip_address is None:
-        local_ip_address = socket.gethostbyname(socket.gethostname())
-    if local_ip_address.startswith('127.'):
-        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
-        s.connect(('8.8.8.8', 53))  # connecting to a UDP address doesn't send packets
-        local_ip_address = s.getsockname()[0]
+        with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
+            s.connect(('8.8.8.8', 53))  # connecting to a UDP address doesn't send packets
+            local_ip_address = s.getsockname()[0]
+
     address = local_ip_address.split('.')
     cs = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
     cs.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
@@ -136,10 +170,10 @@ def discover(timeout=None, local_ip_address=None, discover_ip_address='255.255.2
         responsepacket = bytearray(response[0])
         host = response[1]
         devtype = responsepacket[0x34] | responsepacket[0x35] << 8
-        mac = responsepacket[0x3a:0x40]
+        mac = responsepacket[0x3f:0x39:-1]
         name = responsepacket[0x40:].split(b'\x00')[0].decode('utf-8')
-        cloud = bool(responsepacket[-1])
-        device = gendevice(devtype, host, mac, name=name, cloud=cloud)
+        is_locked = bool(responsepacket[-1])
+        device = gendevice(devtype, host, mac, name=name, is_locked=is_locked)
         cs.close()
         return device
 
@@ -153,23 +187,35 @@ def discover(timeout=None, local_ip_address=None, discover_ip_address='255.255.2
         responsepacket = bytearray(response[0])
         host = response[1]
         devtype = responsepacket[0x34] | responsepacket[0x35] << 8
-        mac = responsepacket[0x3a:0x40]
+        mac = responsepacket[0x3f:0x39:-1]
         name = responsepacket[0x40:].split(b'\x00')[0].decode('utf-8')
-        cloud = bool(responsepacket[-1])
-        device = gendevice(devtype, host, mac, name=name, cloud=cloud)
+        is_locked = bool(responsepacket[-1])
+        device = gendevice(devtype, host, mac, name=name, is_locked=is_locked)
         devices.append(device)
     cs.close()
     return devices
 
 
 class device:
-    def __init__(self, host, mac, devtype, timeout=10, name=None, cloud=None):
+    def __init__(
+        self,
+        host,
+        mac,
+        devtype,
+        timeout=10,
+        name=None,
+        model=None,
+        manufacturer=None,
+        is_locked=None
+    ):
         self.host = host
         self.mac = mac.encode() if isinstance(mac, str) else mac
         self.devtype = devtype if devtype is not None else 0x272a
-        self.name = name
-        self.cloud = cloud
         self.timeout = timeout
+        self.name = name
+        self.model = model
+        self.manufacturer = manufacturer
+        self.is_locked = is_locked
         self.count = random.randrange(0xffff)
         self.iv = bytearray(
             [0x56, 0x2e, 0x17, 0x99, 0x6d, 0x09, 0x3d, 0x28, 0xdd, 0xb3, 0xba, 0x69, 0x5a, 0x2e, 0x6f, 0x58])
@@ -229,11 +275,36 @@ class device:
         if len(key) % 16 != 0:
             return False
 
-        self.id = payload[0x00:0x04]
+        self.id = payload[0x03::-1]
         self.update_aes(key)
 
         return True
 
+    def get_fwversion(self):
+        packet = bytearray([0x68])
+        response = self.send_packet(0x6a, packet)
+        check_error(response[0x22:0x24])
+        payload = self.decrypt(response[0x38:])
+        return payload[0x4] | payload[0x5] << 8        
+
+    def set_name(self, name):
+        packet = bytearray(4)
+        packet += name.encode('utf-8')
+        packet += bytearray(0x50 - len(packet))
+        packet[0x43] = self.is_locked
+        response = self.send_packet(0x6a, packet)
+        check_error(response[0x22:0x24])
+        self.name = name
+
+    def set_lock(self, state):
+        packet = bytearray(4)
+        packet += self.name.encode('utf-8')
+        packet += bytearray(0x50 - len(packet))
+        packet[0x43] = state
+        response = self.send_packet(0x6a, packet)
+        check_error(response[0x22:0x24])
+        self.is_locked = bool(state)
+
     def get_type(self):
         return self.type
 
@@ -253,16 +324,16 @@ class device:
         packet[0x26] = command
         packet[0x28] = self.count & 0xff
         packet[0x29] = self.count >> 8
-        packet[0x2a] = self.mac[0]
-        packet[0x2b] = self.mac[1]
-        packet[0x2c] = self.mac[2]
-        packet[0x2d] = self.mac[3]
-        packet[0x2e] = self.mac[4]
-        packet[0x2f] = self.mac[5]
-        packet[0x30] = self.id[0]
-        packet[0x31] = self.id[1]
-        packet[0x32] = self.id[2]
-        packet[0x33] = self.id[3]
+        packet[0x2a] = self.mac[5]
+        packet[0x2b] = self.mac[4]
+        packet[0x2c] = self.mac[3]
+        packet[0x2d] = self.mac[2]
+        packet[0x2e] = self.mac[1]
+        packet[0x2f] = self.mac[0]
+        packet[0x30] = self.id[3]
+        packet[0x31] = self.id[2]
+        packet[0x32] = self.id[1]
+        packet[0x33] = self.id[0]
 
         # pad the payload for AES encryption
         if payload:
@@ -511,79 +582,39 @@ class sp2(device):
 
 
 class a1(device):
+
+    _SENSORS_AND_LEVELS = (
+        ('light', ('dark', 'dim', 'normal', 'bright')),
+        ('air_quality', ('excellent', 'good', 'normal', 'bad')),
+        ('noise', ('quiet', 'normal', 'noisy')),
+    )
+
     def __init__(self, *args, **kwargs):
         device.__init__(self, *args, **kwargs)
         self.type = "A1"
 
     def check_sensors(self):
-        packet = bytearray(16)
-        packet[0] = 1
-        response = self.send_packet(0x6a, packet)
-        check_error(response[0x22:0x24])
-        data = {}
-        payload = self.decrypt(bytes(response[0x38:]))
-        if isinstance(payload[0x4], int):
-            data['temperature'] = (payload[0x4] * 10 + payload[0x5]) / 10.0
-            data['humidity'] = (payload[0x6] * 10 + payload[0x7]) / 10.0
-            light = payload[0x8]
-            air_quality = payload[0x0a]
-            noise = payload[0xc]
-        else:
-            data['temperature'] = (ord(payload[0x4]) * 10 + ord(payload[0x5])) / 10.0
-            data['humidity'] = (ord(payload[0x6]) * 10 + ord(payload[0x7])) / 10.0
-            light = ord(payload[0x8])
-            air_quality = ord(payload[0x0a])
-            noise = ord(payload[0xc])
-        if light == 0:
-            data['light'] = 'dark'
-        elif light == 1:
-            data['light'] = 'dim'
-        elif light == 2:
-            data['light'] = 'normal'
-        elif light == 3:
-            data['light'] = 'bright'
-        else:
-            data['light'] = 'unknown'
-        if air_quality == 0:
-            data['air_quality'] = 'excellent'
-        elif air_quality == 1:
-            data['air_quality'] = 'good'
-        elif air_quality == 2:
-            data['air_quality'] = 'normal'
-        elif air_quality == 3:
-            data['air_quality'] = 'bad'
-        else:
-            data['air_quality'] = 'unknown'
-        if noise == 0:
-            data['noise'] = 'quiet'
-        elif noise == 1:
-            data['noise'] = 'normal'
-        elif noise == 2:
-            data['noise'] = 'noisy'
-        else:
-            data['noise'] = 'unknown'
+        data = self.check_sensors_raw()
+        for sensor, levels in self._SENSORS_AND_LEVELS:
+            try:
+                data[sensor] = levels[data[sensor]]
+            except IndexError:
+                data[sensor] = 'unknown'
         return data
 
     def check_sensors_raw(self):
-        packet = bytearray(16)
-        packet[0] = 1
+        packet = bytearray([0x1])
         response = self.send_packet(0x6a, packet)
         check_error(response[0x22:0x24])
-        data = {}
         payload = self.decrypt(bytes(response[0x38:]))
-        if isinstance(payload[0x4], int):
-            data['temperature'] = (payload[0x4] * 10 + payload[0x5]) / 10.0
-            data['humidity'] = (payload[0x6] * 10 + payload[0x7]) / 10.0
-            data['light'] = payload[0x8]
-            data['air_quality'] = payload[0x0a]
-            data['noise'] = payload[0xc]
-        else:
-            data['temperature'] = (ord(payload[0x4]) * 10 + ord(payload[0x5])) / 10.0
-            data['humidity'] = (ord(payload[0x6]) * 10 + ord(payload[0x7])) / 10.0
-            data['light'] = ord(payload[0x8])
-            data['air_quality'] = ord(payload[0x0a])
-            data['noise'] = ord(payload[0xc])
-        return data
+        data = bytearray(payload[0x4:])
+        return {
+            'temperature': data[0x0] + data[0x1] / 10.0,
+            'humidity': data[0x2] + data[0x3] / 10.0,
+            'light': data[0x4],
+            'air_quality': data[0x6],
+            'noise': data[0x8],
+        }
 
 
 class rm(device):
@@ -603,7 +634,7 @@ class rm(device):
 
     def send_data(self, data):
         packet = bytearray(self._code_sending_header)
-        packet += bytes([0x02, 0x00, 0x00, 0x00])
+        packet += bytearray([0x02, 0x00, 0x00, 0x00])
         packet += data
         response = self.send_packet(0x6a, packet)
         check_error(response[0x22:0x24])
@@ -646,21 +677,22 @@ class rm(device):
             return True
         return False
 
-    def _read_sensor(self, type, offset, divider):
+    def _check_sensors(self, command):
         packet = bytearray(self._request_header)
-        packet.append(type)
+        packet.append(command)
         response = self.send_packet(0x6a, packet)
         check_error(response[0x22:0x24])
         payload = self.decrypt(bytes(response[0x38:]))
-        value_pos = len(self._request_header) + offset
-        if isinstance(payload[value_pos], int):
-            value = (payload[value_pos] + payload[value_pos+1] / divider)
-        else:
-            value = (ord(payload[value_pos]) + ord(payload[value_pos+1]) / divider)
-        return value
+        return bytearray(payload[len(self._request_header) + 4:])
 
     def check_temperature(self):
-        return self._read_sensor( 0x01, 4, 10.0 )
+        data = self._check_sensors(0x1)
+        return data[0x0] + data[0x1] / 10.0
+
+    def check_sensors(self):
+        data = self._check_sensors(0x1)
+        return {'temperature': data[0x0] + data[0x1] / 10.0}
+
 
 class rm4(rm):
     def __init__(self, *args, **kwargs):
@@ -670,17 +702,21 @@ class rm4(rm):
         self._code_sending_header = b'\xd0\x00'
 
     def check_temperature(self):
-        return self._read_sensor( 0x24, 4, 100.0 )
+        data = self._check_sensors(0x24)
+        return data[0x0] + data[0x1] / 100.0
 
     def check_humidity(self):
-        return self._read_sensor( 0x24, 6, 100.0 )
+        data = self._check_sensors(0x24)
+        return data[0x2] + data[0x3] / 100.0
 
     def check_sensors(self):
+        data = self._check_sensors(0x24)
         return {
-            'temperature': self.check_temperature(),
-            'humidity': self.check_humidity()
+            'temperature': data[0x0] + data[0x1] / 100.0,
+            'humidity': data[0x2] + data[0x3] / 100.0
         }
 
+
 # For legacy compatibility - don't use this
 class rm2(rm):
     def __init__(self):
@@ -1009,8 +1045,8 @@ class lb1(device):
                         'color jumping' : 6,
                         'multicolor jumping' : 7 }
 
-    def __init__(self, host, mac, devtype):
-        device.__init__(self, host, mac, devtype)
+    def __init__(self, *args, **kwargs):
+        device.__init__(self, *args, **kwargs)
         self.type = "SmartBulb"
 
     def send_command(self,command, type = 'set'):
